import functools
from collections import defaultdict
from typing import TYPE_CHECKING, Dict, List, Optional, Sequence

import PySide6QtAds as QtAds
from bidict import bidict
from PySide6.QtWidgets import QSizePolicy

from .views.view import ViewStatePublisherMixin

if TYPE_CHECKING:
    from .views.view import BaseView


class ViewManager:
    """
    Manages views.
    """

    DOCKING_POSITIONS = {
        "center": QtAds.CenterDockWidgetArea,
        "left": QtAds.LeftDockWidgetArea,
        "right": QtAds.RightDockWidgetArea,
        "top": QtAds.TopDockWidgetArea,
        "bottom": QtAds.BottomDockWidgetArea,
    }

    def __init__(self, workspace):
        self.workspace = workspace
        self.views: List[BaseView] = []
        self.docks = []
        self.view_to_dock = bidict()
        self.views_by_category: Dict[str, List[BaseView]] = defaultdict(list)
        self.views_by_activation: List[BaseView] = []
        self.main_window.dock_manager.focusedDockWidgetChanged.connect(self._on_dock_widget_focus_changed)

    @property
    def main_window(self):
        return self.workspace._main_window

    def main_window_initialized(self) -> None:
        """
        Invoked by the main window after it has finished initialization. Views can override
        BaseView.mainWindowInitializedEvent() to support delayed initialization or loading.
        """
        for view in self.views:
            view.mainWindowInitializedEvent()

    def _update_view_index_in_category(self, view: "BaseView") -> None:
        """
        Set lowest available index value in category for a view not yet added.
        """
        existing_views = self.views_by_category[view.category]
        if view in existing_views:
            return
        existing_ids = {view.index for view in existing_views}
        max_id = max(existing_ids) if existing_ids else 0
        candidates = set(range(1, max_id + 2)) - existing_ids
        view.index = min(candidates)

    def _promote_view(self, view: "BaseView") -> None:
        """
        Move view to first position in views_by_activation.
        """
        self.views_by_activation.remove(view)
        self.views_by_activation.insert(0, view)

    def add_view(self, view: "BaseView") -> None:
        """
        Add a view to this workspace.

        :param view:            The view to add.
        :return:                None
        """
        self._update_view_index_in_category(view)
        self.views_by_category[view.category].append(view)
        self.views_by_activation.insert(0, view)

        dw = QtAds.CDockWidget(view.caption)
        dw.setFeature(QtAds.CDockWidget.DockWidgetDeleteOnClose, True)
        dw.closed.connect(functools.partial(self._on_dock_widget_closed, dw))
        dw.setWidget(view)

        area = self.DOCKING_POSITIONS.get(view.default_docking_position, QtAds.RightDockWidgetArea)
        area_widget = self.main_window.dock_manager.addDockWidgetTab(area, dw)
        self.main_window.init_shortcuts_on_dock(dw)

        self.views.append(view)
        self.docks.append(dw)
        self.view_to_dock[view] = dw

        if view.default_docking_position == "center":
            policy = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
            policy.setHorizontalStretch(1)
            policy.setVerticalStretch(1)
            area_widget.setSizePolicy(policy)

    @property
    def most_recently_focused_view(self) -> Optional["BaseView"]:
        if self.views_by_activation:
            return self.views_by_activation[0]
        return None

    def get_most_recently_focused_view_by_docking_area(self, area: str) -> Optional["BaseView"]:
        for view in self.views_by_activation:
            if view.default_docking_position == area:
                return view
        return None

    def _on_dock_widget_focus_changed(self, _, new: Optional[QtAds.CDockWidget]) -> None:
        """
        Handle dock focus events.
        """
        view = self.view_to_dock.inverse.get(new, None)
        if view:
            self._promote_view(view)
        if isinstance(view, ViewStatePublisherMixin):
            view.on_focused()

    def _on_dock_widget_closed(self, dock: QtAds.CDockWidget) -> None:
        """
        Handle dock widget close event.
        """
        if dock not in self.docks:
            return

        self.docks.remove(dock)
        view = self.view_to_dock.inverse.pop(dock, None)
        if view:
            view.close()
            self.remove_view(view)

    def remove_view(self, view: "BaseView") -> None:
        """
        Remove a view from this workspace
        """
        if view not in self.views:
            return

        self.views.remove(view)
        self.views_by_category[view.category].remove(view)
        dock = self.view_to_dock.pop(view, None)
        if dock:
            dock.closeDockWidget()
        self.views_by_activation.remove(view)

    def raise_view(self, view: "BaseView") -> None:
        """
        Find the dock widget of a view, and then bring that dock widget to front.
        """
        self._promote_view(view)

        # find the dock widget by the view
        dock = self.view_to_dock.get(view, None)
        if dock is None:
            return

        if not dock.isTabbed():
            dock.show()
        dock.raise_()
        view.focusWidget()

    def get_center_docks(self) -> Sequence[QtAds.CDockWidget]:
        """
        Get the center dockable views
        """
        return [
            dock
            for dock in self.docks
            if dock.widget() is not None and dock.widget().default_docking_position == "center"
        ]

    def first_view_in_category(self, category: str) -> Optional["BaseView"]:
        """
        Return the first view in a specific category.
        """
        if self.views_by_category[category]:
            return self.views_by_category[category][0]
        return None

    def current_view_in_category(self, category: str) -> Optional["BaseView"]:
        """
        Return the current in a specific category.
        """
        for view in self.views_by_activation:
            if view.category == category:
                return view
        return None

    def _adjust_current_tab_idx(self, offset: int) -> None:
        """
        Select tab in same dock area with index equal to index of most recently activated center view plus `offset`.
        """
        view = self.get_most_recently_focused_view_by_docking_area("center")
        if view is None:
            return
        area = self.view_to_dock[view].dockAreaWidget()
        idx = area.currentIndex()
        if idx < 0:
            return
        area.setCurrentIndex((idx + offset) % area.dockWidgetsCount())

    def next_tab(self) -> None:
        """
        Shift to the next tab
        """
        self._adjust_current_tab_idx(1)

    def previous_tab(self) -> None:
        """
        Shift to the previous tab
        """
        self._adjust_current_tab_idx(-1)

    @property
    def current_tab(self) -> Optional["BaseView"]:
        return self.get_most_recently_focused_view_by_docking_area("center")

    def _handle_raise_view(self, view: "BaseView") -> None:
        self.workspace.plugins.handle_raise_view(view)



from PySide6.QtCore import QPointF, Qt
from PySide6.QtGui import QBrush, QColor, QPen

from angrmanagement.utils.graph_layouter import GraphLayouter

from .qgraph import QZoomableDraggableGraphicsView


class QSymExecGraph(QZoomableDraggableGraphicsView):
    LEFT_PADDING = 2000
    TOP_PADDING = 2000

    def __init__(self, current_state, workspace, symexec_view, parent=None):
        super().__init__(parent=parent)

        self.state = current_state
        self._symexec_view = symexec_view

        self._graph = None
        self.blocks = set()
        self._edges = []

        self._edge_paths = []

        self.state.am_subscribe(self._watch_state)

    @property
    def graph(self):
        return self._graph

    @graph.setter
    def graph(self, v):
        if v is not self._graph:
            self._graph = v
            self.reload()

    def reload(self):
        self.request_relayout()

    def request_relayout(self):
        self._reset_scene()
        if self.graph is None:
            return

        # remove all edges
        scene = self.scene()
        for p in self._edge_paths:
            scene.removeItem(p)

        # remove all nodes
        self.blocks.clear()
        # self.remove_all_children()
        self._edge_paths = []

        node_sizes = {}
        for node in self.graph.nodes():
            self.blocks.add(node)
            node_sizes[node] = (node.width, node.height)
        gl = GraphLayouter(self.graph, node_sizes, node_compare_key=lambda n: 0)

        self._edges = gl.edges

        min_x, max_x, min_y, max_y = 0, 0, 0, 0

        scene = self.scene()
        for node, (x, y) in gl.node_coordinates.items():
            scene.addItem(node)
            node.setPos(x, y)
            min_x = min(min_x, node.x())
            max_x = max(max_x, node.x() + node.width)
            min_y = min(min_y, node.y())
            max_y = max(max_y, node.y() + node.height)

        min_x -= self.LEFT_PADDING
        max_x += self.LEFT_PADDING
        min_y -= self.TOP_PADDING
        max_y += self.TOP_PADDING
        (max_x - min_x) + 2 * self.LEFT_PADDING
        (max_y - min_y) + 2 * self.TOP_PADDING

        self._reset_view()

    #
    # Event handlers
    #

    def keyPressEvent(self, event):
        """

        :param QKeyEvent event:
        :return:
        """

        key = event.key()

        if key == Qt.Key_Tab:
            self._symexec_view.switch_to_disassembly_view()
            event.accept()

        super().keyPressEvent(event)

    def _watch_state(self, **kwargs):
        for block in self.blocks:
            if block.get_state() == self.state:
                block.selected = True
            else:
                block.selected = False

        self.viewport().update()

    #
    # Private methods
    #

    def _initial_position(self):
        ibr = self.scene().itemsBoundingRect()
        return ibr.center()

    def _block_from_state(self, state):
        for block in self.blocks:
            if block.get_state() == state:
                return block
        return None

    def _draw_edges(self, painter, topleft_point, bottomright_point):
        for edge in self._edges:
            edge_coords = edge.coordinates

            color = QColor(0x70, 0x70, 0x70)
            pen = QPen(color)
            pen.setWidth(1.5)
            painter.setPen(pen)

            for from_, to_ in zip(edge_coords, edge_coords[1:]):
                start_point = QPointF(*from_)
                end_point = QPointF(*to_)
                # optimization: don't draw edges that are outside of the current scope
                if (start_point.x() > bottomright_point.x() or start_point.y() > bottomright_point.y()) and (
                    end_point.x() > bottomright_point.x() or end_point.y() > bottomright_point.y()
                ):
                    continue
                elif (start_point.x() < topleft_point.x() or start_point.y() < topleft_point.y()) and (
                    end_point.x() < topleft_point.x() or end_point.y() < topleft_point.y()
                ):
                    continue
                painter.drawPolyline((start_point, end_point))

            # arrow
            # end_point = self.mapToScene(*edges[-1])
            end_point = (edge_coords[-1][0], edge_coords[-1][1])
            arrow = [
                QPointF(end_point[0] - 3, end_point[1]),
                QPointF(end_point[0] + 3, end_point[1]),
                QPointF(end_point[0], end_point[1] + 6),
            ]
            brush = QBrush(color)
            painter.setBrush(brush)
            painter.drawPolygon(arrow)



from typing import TYPE_CHECKING, Iterable

if TYPE_CHECKING:
    from PySide6.QtWidgets import QGridLayout, QWidget


def add_to_grid(lyt: "QGridLayout", cols: int, widgets: Iterable["QWidget"]):
    """
    Adds widgets to a grid layout given a desired column count.
    """
    r = lyt.rowCount()
    c = 0
    for item in widgets:
        lyt.addWidget(item, r, c)
        c += 1
        if c >= cols:
            c = 0
            r += 1



import os
from typing import TYPE_CHECKING, Any, Callable

from PySide6.QtCore import QSize, Signal
from PySide6.QtGui import QAction, QIcon, QMouseEvent
from PySide6.QtWidgets import QMenu, QStyle, QToolBar, QToolButton

from angrmanagement.config import IMG_LOCATION

from .toolbar import Toolbar

if TYPE_CHECKING:
    from angrmanagement.logic.disassembly import JumpHistory


class NavToolButton(QToolButton):
    """
    Widget to allow navigating a JumpHistory stack
    """

    triggered = Signal()
    triggeredFromMenu = Signal(int)

    def __init__(self, jump_history: "JumpHistory", direction_forward: bool = False, parent=None):
        super().__init__(parent)
        self._dir_fwd = direction_forward
        self._jump_history = jump_history
        self.setPopupMode(QToolButton.MenuButtonPopup)
        self._init_menu()

        if direction_forward:
            lbl = "Forward"
            ico = QIcon(os.path.join(IMG_LOCATION, "toolbar-forward.png"))
        else:
            lbl = "Back"
            ico = QIcon(os.path.join(IMG_LOCATION, "toolbar-previous.png"))

        a = QAction(ico, lbl, self)
        a.triggered.connect(self._on_button_activated)
        self.setDefaultAction(a)

    def _on_button_activated(self):
        self.triggered.emit()

    def _on_menu_action_activated(self, checked):  # pylint:disable=unused-argument
        pos = self.sender().data()
        self.triggeredFromMenu.emit(pos)

    def _init_menu(self):
        self._menu = QMenu()
        pos = self._jump_history.pos
        if pos < 0:
            pos += len(self._jump_history.history)

        actions = []
        for i, point in enumerate(self._jump_history.history):
            a = QAction(f"{i}: {point:x}", self)
            a.setData(i)
            a.setCheckable(True)
            a.setChecked(pos == i)
            a.triggered.connect(self._on_menu_action_activated)
            actions.append(a)

        actions.reverse()
        self._menu.addActions(actions)
        self.setMenu(self._menu)

    def mousePressEvent(self, e: QMouseEvent):
        self._init_menu()
        super().mousePressEvent(e)


class NavToolbar(Toolbar):
    """
    Navigation toolbar with forward, back, and menu-based navigation of a JumpHistory stack
    """

    def __init__(
        self,
        jump_history: "JumpHistory",
        back_triggered: Callable[[], Any],
        forward_triggered: Callable[[], Any],
        point_triggered: Callable[[int], Any],
        small_icon: bool,
        window,
    ):
        super().__init__(window, "Navigation")
        self._jump_history = jump_history
        self._back_triggered = back_triggered
        self._forward_triggered = forward_triggered
        self._point_triggered = point_triggered
        self._small_icon = small_icon

    def qtoolbar(self) -> QToolBar:
        tb = QToolBar(self.window)
        if self._small_icon:
            sm_ico_pm = tb.style().pixelMetric(QStyle.PM_SmallIconSize, None, tb)
            tb.setIconSize(QSize(sm_ico_pm, sm_ico_pm))

        back_btn = NavToolButton(self._jump_history, False, tb)
        back_btn.triggered.connect(self._back_triggered)
        back_btn.triggeredFromMenu.connect(self._point_triggered)
        tb.addWidget(back_btn)

        fwd_btn = NavToolButton(self._jump_history, True, tb)
        fwd_btn.triggered.connect(self._forward_triggered)
        fwd_btn.triggeredFromMenu.connect(self._point_triggered)
        tb.addWidget(fwd_btn)

        return tb



from typing import TYPE_CHECKING, Dict, Optional, Set

import networkx
from angr.analyses.proximity_graph import (
    BaseProxiNode,
    CallProxiNode,
    FunctionProxiNode,
    StringProxiNode,
    VariableProxiNode,
)
from PySide6.QtCore import QSize
from PySide6.QtWidgets import QHBoxLayout

from angrmanagement.ui.views.view import BaseView
from angrmanagement.ui.widgets.qproximity_graph import QProximityGraph
from angrmanagement.ui.widgets.qproximitygraph_block import (
    QProximityGraphBlock,
    QProximityGraphCallBlock,
    QProximityGraphFunctionBlock,
    QProximityGraphStringBlock,
)

if TYPE_CHECKING:
    from angr.knowledge_plugins.functions import Function


class ProximityView(BaseView):
    """
    Proximity View
    """

    def __init__(self, workspace, instance, default_docking_position, *args, **kwargs):
        super().__init__("proximity", workspace, instance, default_docking_position, *args, **kwargs)

        self.base_caption = "Proximity"

        self._function: Optional[Function] = None
        self._expand_function_addrs: Set[int] = set()

        # UI widgets
        self._graph_widget: QProximityGraph = None

        # data
        self._proximity_graph: Optional[networkx.DiGraph] = None  # generated by ProximityGraphAnalysis
        self._graph: Optional[networkx.DiGraph] = None
        self.hovered_block: Optional[QProximityGraphBlock] = None

        self._init_widgets()
        self._register_events()

    @property
    def function(self):
        return self._function

    @function.setter
    def function(self, v):
        if v is not None and v is not self._function:
            self._function = v
            self._expand_function_addrs.clear()
            self.run_analysis()

    def get_decompilation(self):
        # FIXME: Get rid of this fucker and replace it with the new decompilation manager
        inst = self.instance
        dec = inst.project.analyses.Decompiler(
            self.function,
            cfg=self.instance.cfg,
        )
        return dec

    def run_analysis(self):
        dec = self.get_decompilation()

        inst = self.instance
        prox = inst.project.analyses.Proximity(
            self.function,
            inst.cfg,
            inst.kb.xrefs,
            decompilation=dec,
            expand_funcs=self._expand_function_addrs,
        )
        self._proximity_graph = prox.graph

        self.reload()

    def hover_enter_block(self, block: QProximityGraphBlock):
        self.hovered_block = block
        if self._graph_widget is not None:
            self._graph_widget.on_block_hovered(block)
        self.redraw_graph()

    def hover_leave_block(self, block: QProximityGraphBlock):  # pylint: disable=unused-argument
        self.hovered_block = None
        self.redraw_graph()

    def expand_function(self, func):
        if func.addr not in self._expand_function_addrs:
            self._expand_function_addrs.add(func.addr)
            # re-run the analysis
            self.run_analysis()

    def collapse_function(self, func):
        if func.addr in self._expand_function_addrs:
            self._expand_function_addrs.discard(func.addr)
            # re-run the analysis
            self.run_analysis()

    def on_screen_changed(self):
        if self._graph_widget is not None:
            self._graph_widget.refresh()

    def reload(self):
        if self._graph_widget is None:
            return
        # re-generate the graph
        if not self._proximity_graph:
            self._graph = None
            self._graph_widget.graph = None
            self._graph_widget.request_relayout()
            return

        self._graph = self._create_ui_graph()
        self._graph_widget.graph = self._graph

    def clear(self):
        self._proximity_graph = None
        self.reload()

    def redraw_graph(self):
        if self._graph_widget is not None:
            self._graph_widget.viewport().update()

    def sizeHint(self):
        return QSize(400, 800)

    def _init_widgets(self):
        self._graph_widget = QProximityGraph(self.instance, self)

        hlayout = QHBoxLayout()
        hlayout.addWidget(self._graph_widget)
        hlayout.setContentsMargins(0, 0, 0, 0)

        self.setLayout(hlayout)

    def _register_events(self):
        self.workspace.current_screen.am_subscribe(self.on_screen_changed)

    def _unregister_events(self):
        self.workspace.current_screen.am_unsubscribe(self.on_screen_changed)

    def closeEvent(self, event):
        self._unregister_events()
        super().closeEvent(event)

    def _convert_node(
        self, node: BaseProxiNode, converted: Dict[BaseProxiNode, QProximityGraphBlock]
    ) -> Optional[QProximityGraphBlock]:
        if node in converted:
            return converted[node]

        if isinstance(node, StringProxiNode):
            new_node = QProximityGraphStringBlock(False, self, node)
        elif isinstance(node, FunctionProxiNode):
            new_node = QProximityGraphFunctionBlock(False, self, node)
        elif isinstance(node, CallProxiNode):
            new_node = QProximityGraphCallBlock(False, self, node)
        elif isinstance(node, VariableProxiNode):
            new_node = QProximityGraphBlock
        elif isinstance(node, BaseProxiNode):
            new_node = QProximityGraphBlock(False, self, node)
        else:
            raise TypeError(f"Unsupported type of proximity node {type(node)}.")
        converted[node] = new_node
        return new_node

    def _create_ui_graph(self) -> networkx.DiGraph:
        g = networkx.DiGraph()

        converted = {}
        for proxi_node in self._proximity_graph.nodes():
            node = self._convert_node(proxi_node, converted)
            if node is not None:
                g.add_node(node)

        for n0, n1 in self._proximity_graph.edges():
            n0_ = self._convert_node(n0, converted)
            n1_ = self._convert_node(n1, converted)

            g.add_edge(n0_, n1_)

        return g



from typing import TYPE_CHECKING, Any, Optional

from PySide6.QtCore import QAbstractTableModel, QSize, Qt
from PySide6.QtWidgets import QAbstractItemView, QHeaderView, QMenu, QTableView, QVBoxLayout

from .view import BaseView

if TYPE_CHECKING:
    import PySide6

    from angrmanagement.data.trace import Trace


class QTraceTableModel(QAbstractTableModel):
    """
    Trace table model.
    """

    Headers = ["Source"]
    COL_SOURCE = 0

    def __init__(self, instance):
        super().__init__()
        self.instance = instance
        self.instance.current_trace.am_subscribe(self._on_traces_updated)
        self.instance.traces.am_subscribe(self._on_traces_updated)

    def shutdown(self):
        self.instance.current_trace.am_unsubscribe(self._on_traces_updated)
        self.instance.traces.am_unsubscribe(self._on_traces_updated)

    def _on_traces_updated(self, **kwargs):  # pylint:disable=unused-argument
        self.beginResetModel()
        self.endResetModel()

    def rowCount(self, parent: "PySide6.QtCore.QModelIndex" = ...) -> int:  # pylint:disable=unused-argument
        return len(self.instance.traces)

    def columnCount(self, parent: "PySide6.QtCore.QModelIndex" = ...) -> int:  # pylint:disable=unused-argument
        return len(self.Headers)

    def headerData(
        self, section: int, orientation: "PySide6.QtCore.Qt.Orientation", role: int = ...
    ) -> Any:  # pylint:disable=unused-argument
        if role != Qt.DisplayRole:
            return None
        if section < len(self.Headers):
            return self.Headers[section]
        return None

    def data(self, index: "PySide6.QtCore.QModelIndex", role: int = ...) -> Any:
        if not index.isValid():
            return None
        row = index.row()
        if row >= len(self.instance.traces):
            return None
        col = index.column()
        if role == Qt.DisplayRole:
            return self._get_column_text(self.instance.traces[row], col)
        else:
            return None

    def _get_column_text(self, trace: "Trace", col: int) -> Any:
        if col == QTraceTableModel.COL_SOURCE:
            return trace.source + (" (Current)" if self.instance.current_trace.am_obj is trace else "")
        else:
            return None


class QTraceTableWidget(QTableView):
    """
    Trace table widget.
    """

    def __init__(self, instance, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.instance = instance

        hheader = self.horizontalHeader()
        hheader.setVisible(True)
        hheader.setStretchLastSection(True)

        vheader = self.verticalHeader()
        vheader.setVisible(False)
        vheader.setDefaultSectionSize(20)

        self.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.setHorizontalScrollMode(QAbstractItemView.ScrollPerPixel)

        self.model: QTraceTableModel = QTraceTableModel(instance)
        self.setModel(self.model)

        hheader.setSectionResizeMode(0, QHeaderView.ResizeToContents)

    #
    # Events
    #

    def closeEvent(self, event):
        self.model.shutdown()
        super().closeEvent(event)

    def contextMenuEvent(self, event):
        selected_rows = {i.row() for i in self.selectedIndexes()}
        traces = [self.instance.traces[r] for r in selected_rows]
        if len(traces):
            menu = QMenu("", self)
            if len(traces) == 1 and not self.workspace(traces[0]):
                menu.addAction("Use as current trace", lambda: self.workspace.set_current_trace(traces[0]))

            def remove_selected_traces():
                for t in traces:
                    self.workspace.remove_trace(t)

            menu.addAction("Remove trace" + ("s" if len(traces) > 1 else ""), remove_selected_traces)
            menu.exec_(event.globalPos())


class TracesView(BaseView):
    """
    Traces table view.
    """

    def __init__(self, workspace, instance, default_docking_position, *args, **kwargs):
        super().__init__("traces", workspace, instance, default_docking_position, *args, **kwargs)

        self.base_caption = "Traces"
        self._tbl_widget: Optional[QTraceTableWidget] = None
        self._init_widgets()
        self.reload()

    def reload(self):
        pass

    @staticmethod
    def minimumSizeHint(*args, **kwargs):  # pylint:disable=unused-argument
        return QSize(200, 200)

    def _init_widgets(self):
        vlayout = QVBoxLayout()
        vlayout.setContentsMargins(0, 0, 0, 0)
        self._tbl_widget = QTraceTableWidget(self.instance, self)
        vlayout.addWidget(self._tbl_widget)
        self.setLayout(vlayout)



# pylint:disable=missing-class-docstring,wrong-import-order,no-self-use
import os
import sys
import threading
import unittest

import angr
from common import start_main_window_and_event_loop, test_location


class TestWorkflow(unittest.TestCase):
    def setUp(self):
        self.event = threading.Event()
        _, self.main = start_main_window_and_event_loop(self.event)

    def tearDown(self) -> None:
        self.event.set()
        del self.main

    def test_workflow(self):
        main = self.main
        proj = angr.Project(os.path.join(test_location, "x86_64", "true"), auto_load_libs=False)
        main.workspace.main_instance.project.am_obj = proj
        main.workspace.main_instance.project.am_event()
        main.workspace.main_instance.join_all_jobs()


if __name__ == "__main__":
    unittest.main(argv=sys.argv)



from typing import TYPE_CHECKING

from PySide6.QtCore import QObject, Signal

from angrmanagement.data.object_container import ObjectContainer

if TYPE_CHECKING:
    from angr.sim_variable import SimVariable


class OperandHighlightMode:
    SAME_IDENT = 0
    SAME_TEXT = 1


class OperandDescriptor:
    __slots__ = (
        "text",
        "num_value",
        "func_addr",
        "variable_ident",
    )

    def __init__(self, text, num_value, func_addr=None, variable_ident=None):
        self.text = text
        self.num_value = num_value
        self.func_addr = func_addr
        self.variable_ident = variable_ident


class InfoDock(QObject):
    """
    Stores information associated to a disassembly view. Such information will be shared between the graph view and the
    linear view.
    """

    qblock_code_obj_selection_changed = Signal()

    def __init__(self, disasm_view):
        super().__init__()
        self.disasm_view = disasm_view

        self.induction_variable_analysis = None
        self.variable_manager = None

        self.highlight_mode = OperandHighlightMode.SAME_IDENT  # default highlight mode

        self.selected_insns = ObjectContainer(set(), "The currently selected instructions")
        self.selected_operands = ObjectContainer({}, "The currently selected instruction operands")
        self.selected_blocks = ObjectContainer(set(), "The currently selected blocks")
        self.hovered_block = ObjectContainer(None, "The currently hovered block")
        self.hovered_edge = ObjectContainer(None, "The currently hovered edge")
        self.selected_labels = ObjectContainer(set(), "The currently selected labels")
        self.selected_variables = ObjectContainer(set(), "The currently selected variables")
        self.selected_qblock_code_obj = None

    @property
    def smart_highlighting(self):
        return self.highlight_mode == OperandHighlightMode.SAME_IDENT

    @smart_highlighting.setter
    def smart_highlighting(self, v):
        if v:
            self.highlight_mode = OperandHighlightMode.SAME_IDENT
        else:
            self.highlight_mode = OperandHighlightMode.SAME_TEXT

    def initialize(self):
        self.selected_blocks.clear()
        self.selected_insns.clear()
        self.selected_operands.clear()
        self.hovered_block.am_obj = None

    def copy(self) -> "InfoDock":
        r = InfoDock(self.disasm_view)
        r.variable_manager = self.variable_manager
        r.highlight_mode = self.highlight_mode
        r.selected_insns.am_obj = set(self.selected_insns.am_obj)
        r.selected_operands.am_obj = dict(self.selected_operands.am_obj)
        r.selected_blocks.am_obj = set(self.selected_blocks.am_obj)
        r.hovered_block.am_obj = self.hovered_block.am_obj
        r.hovered_edge.am_obj = self.hovered_edge.am_obj
        r.selected_labels.am_obj = set(self.selected_labels.am_obj)
        return r

    def hover_edge(self, src_addr, dst_addr):
        self.hovered_edge.am_obj = src_addr, dst_addr
        self.hovered_edge.am_event()

    def unhover_edge(self, src_addr, dst_addr):
        if self.hovered_edge.am_obj == (src_addr, dst_addr):
            self.hovered_edge.am_obj = None
            self.hovered_edge.am_event()

    def hover_block(self, block_addr):
        self.hovered_block.am_obj = block_addr
        self.hovered_block.am_event()

    def unhover_block(self, block_addr):
        if self.hovered_block.am_obj == block_addr:
            self.hovered_block.am_obj = None
            self.hovered_block.am_event()

    def clear_hovered_block(self):
        self.hovered_block.am_obj = None
        self.hovered_block.am_event()

    def select_block(self, block_addr):
        self.selected_blocks.clear()  # selecting one block at a time
        self.selected_blocks.add(block_addr)
        self.selected_blocks.am_event()
        self._update_published_view_state()

    def unselect_block(self, block_addr):
        if block_addr in self.selected_blocks:
            self.selected_blocks.remove(block_addr)
            self.selected_blocks.am_event()
        self._update_published_view_state()

    def select_instruction(self, insn_addr, unique=True, insn_pos=None, use_animation=True):
        self.disasm_view.set_synchronized_cursor_address(insn_addr)

        self.unselect_all_labels()
        if insn_addr not in self.selected_insns:
            if unique:
                # unselect existing ones
                self.unselect_all_instructions()
                self.selected_insns.add(insn_addr)
            else:
                self.selected_insns.add(insn_addr)
            self.disasm_view.current_graph.show_instruction(insn_addr, insn_pos=insn_pos, use_animation=use_animation)
            self.selected_insns.am_event(insn_addr=insn_addr)

        self._update_published_view_state()

    def unselect_instruction(self, insn_addr):
        if insn_addr in self.selected_insns:
            self.selected_insns.remove(insn_addr)
            self.selected_insns.am_event()
        self._update_published_view_state()

    def unselect_all_instructions(self):
        if self.selected_insns:
            self.selected_insns.clear()
            self.selected_insns.am_event()
        self._update_published_view_state()

    def select_operand(self, ins_addr: int, operand_index: int, operand: OperandDescriptor, unique: bool = False):
        """
        Mark an operand as selected.

        :param ins_addr:                Address of the instruction.
        :param operand_index:           Index of the operand.
        :param operand:   Data of the operand.
        :param unique:                 If this is a unique selection or not.
        :return:                            None
        """

        tpl = ins_addr, operand_index
        if tpl not in self.selected_operands:
            if unique:
                self.selected_operands.clear()
            self.selected_operands[tpl] = operand
            self.selected_operands.am_event()

    def unselect_operand(self, insn_addr, operand_idx):
        if (insn_addr, operand_idx) in self.selected_operands:
            self.selected_operands.pop((insn_addr, operand_idx))
            self.selected_operands.am_event()

    def unselect_all_operands(self):
        if self.selected_operands:
            self.selected_operands.clear()
            self.selected_operands.am_event()

    def select_label(self, label_addr):
        self.disasm_view.set_synchronized_cursor_address(label_addr)

        # only one label can be selected at a time
        # also, clear selection of instructions and operands
        self.unselect_all_instructions()
        self.unselect_all_operands()

        self.selected_labels.clear()
        self.selected_labels.add(label_addr)
        self.selected_labels.am_event()

        self._update_published_view_state()

    def toggle_label_selection(self, addr: int) -> None:
        """
        Toggle the selection state of a label in the disassembly view.

        :param addr:    Address of the instruction to toggle.
        """

        if addr in self.selected_labels:
            self.unselect_label(addr)
        else:
            self.select_label(addr)

    def unselect_label(self, label_addr):
        if label_addr in self.selected_labels:
            self.selected_labels.remove(label_addr)
            self.selected_labels.am_event()
        self._update_published_view_state()

    def unselect_all_labels(self):
        self.selected_labels.clear()
        self.selected_labels.am_event()
        self._update_published_view_state()

    def toggle_instruction_selection(self, insn_addr, insn_pos=None, unique=False):
        """
        Toggle the selection state of an instruction in the disassembly view.

        :param int insn_addr: Address of the instruction to toggle.
        :return:              None
        """

        if insn_addr in self.selected_insns:
            self.unselect_instruction(insn_addr)
        else:
            self.select_instruction(insn_addr, unique=unique, insn_pos=insn_pos)

    def toggle_operand_selection(self, insn_addr, operand_idx, operand, insn_pos=None, unique=False):
        """
        Toggle the selection state of an operand of an instruction in the disassembly view.

        :param int insn_addr:   Address of the instruction to toggle.
        :param int operand_idx: The operand to toggle.
        :param operand:         The operand instance.
        :return:                True if this operand is now selected, False otherwise.
        :rtype:                 bool
        """

        if (insn_addr, operand_idx) in self.selected_operands:
            self.unselect_operand(insn_addr, operand_idx)
            return False
        else:
            self.select_operand(insn_addr, operand_idx, operand, unique=unique)
            self.disasm_view.current_graph.show_instruction(insn_addr, insn_pos=insn_pos)
            return True

    def select_variable(self, unified_variable: "SimVariable", unique: bool = True):
        self.unselect_all_labels()
        self.unselect_all_instructions()
        self.unselect_all_operands()

        if unique:
            self.selected_variables.clear()
        if unified_variable not in self.selected_variables:
            self.selected_variables.add(unified_variable)
            self.selected_variables.am_event()

    def unselect_variable(self, unified_variable: "SimVariable"):
        if unified_variable in self.selected_variables:
            self.selected_variables.remove(unified_variable)
            self.selected_variables.am_event()

    def toggle_variable_selection(self, unified_variable: "SimVariable", unique: bool = True):
        if len(self.selected_variables) > 1:
            # multiple variables are selected
            if unique:
                # clear existing selections and select this one
                self.select_variable(unified_variable, unique=True)
                return True

        if unified_variable in self.selected_variables:
            self.unselect_variable(unified_variable)
            return False
        else:
            self.select_variable(unified_variable, unique=unique)
            return True

    def clear_selection(self):
        self.selected_blocks.clear()
        self.selected_blocks.am_event()

        self.selected_insns.clear()
        self.selected_insns.am_event()

        self.selected_operands.clear()
        self.selected_operands.am_event()

        self.selected_variables.clear()
        self.selected_variables.am_event()

        self._update_published_view_state()

    def is_edge_hovered(self, src_addr, dst_addr):
        return self.hovered_edge.am_obj == (src_addr, dst_addr)

    def is_block_hovered(self, block_addr):
        return block_addr == self.hovered_block.am_obj

    def is_block_selected(self, block_addr):
        return block_addr in self.selected_blocks

    def is_instruction_selected(self, ins_addr):
        """
        Check if an instruction at @ins_addr is currently selected or not.

        :param int ins_addr:    Address of the instruction.
        :return:                True if it is selected, False otherwise.
        :rtype:                 bool
        """
        return ins_addr in self.selected_insns

    def is_operand_selected(self, ins_addr, operand_index):
        """
        Check if an operand at @ins_addr and @operand_index is currently selected or not.

        :param int ins_addr:        Address of the instruction
        :param int operand_index:   Index of the operand.
        :return:                    bool
        """
        return (ins_addr, operand_index) in self.selected_operands

    def is_label_selected(self, label_addr):
        return label_addr in self.selected_labels

    def is_variable_selected(self, unique_variable_ident: str) -> bool:
        return unique_variable_ident in self.selected_variables

    def should_highlight_operand(self, selected, operand):
        if selected is None:
            return False

        if self.highlight_mode == OperandHighlightMode.SAME_TEXT or selected.variable is None:
            # when there is no related variable, we highlight as long as they have the same text
            return operand.text == selected.text
        elif self.highlight_mode == OperandHighlightMode.SAME_IDENT:
            if selected.variable is not None and operand.variable is not None:
                return selected.variable.ident == operand.variable.ident

        return False

    def select_qblock_code_obj(self, obj):
        """
        For QBlockCodeObj, we simply track selected state for now and handle
        matching in object handlers
        """
        self.selected_qblock_code_obj = obj
        self.qblock_code_obj_selection_changed.emit()

    def _update_published_view_state(self):
        self.disasm_view.published_view_state.cursors = self.selected_insns.union(self.selected_labels)
        self.disasm_view.notify_view_state_updated()



import logging
import math

import networkx as nx
from angr.knowledge_plugins.functions import Function
from PySide6.QtGui import QColor

_l = logging.getLogger(name=__name__)


class AFLQemuBitmap:
    HIT_COLOR = QColor(0xEE, 0xFF, 0xEE)
    MISS_COLOR = QColor(0x99, 0x00, 0x00, 0x30)
    FUNCTION_NOT_VISITED_COLOR = QColor(0x99, 0x00, 0x00, 0x20)
    BUCKET_COLORS = [
        QColor(0xEF, 0x65, 0x48, 0x20),
        QColor(0xFC, 0x8D, 0x59, 0x60),
        QColor(0xFD, 0xBB, 0x84, 0x60),
        QColor(0xFD, 0xD4, 0x9E, 0x60),
    ]

    def __init__(self, workspace, bitmap, base_addr, bits_inverted=False):
        self.workspace = workspace
        self.virgin_bitmap = bitmap
        if bits_inverted:
            # invert all bits
            self.virgin_bitmap = bytes([b ^ 0xFF for b in self.virgin_bitmap])
        self.bitmap_size = len(self.virgin_bitmap)
        assert self.bitmap_size == 1 << (self.bitmap_size.bit_length() - 1)
        self.function_info = {}
        self._hitcount_graphs = {}
        self._node_hitcounts = {}
        self._node_hitcount_summary = {}

        project = self.workspace.main_instance.project
        self.project_baddr = project.loader.main_object.mapped_base
        self.runtime_baddr = base_addr

        self._compute_hitcounts()

    def _compute_hitcounts(self):
        func_manager = self.workspace.main_instance.project.kb.functions
        for func_addr in func_manager:
            func = func_manager[func_addr]

            hitc_g = self._parse_bitmap(func)
            node_hitc = {n.addr: data["hitcount"] for n, data in hitc_g.nodes(data=True)}
            self._hitcount_graphs[func] = hitc_g
            self._node_hitcounts[func] = node_hitc
            for addr, hitcount in node_hitc.items():
                old = self._node_hitcount_summary.get(addr, 0)
                new = max(old, hitcount)
                self._node_hitcount_summary[addr] = new

    def get_hit_miss_color(self, addr):
        # TODO: sometimes there's addresses here that are not in the hitcount, don't know why
        hitcount = self._node_hitcount_summary.get(addr, 0)
        if hitcount == 0:
            return AFLQemuBitmap.MISS_COLOR
        else:
            return AFLQemuBitmap.HIT_COLOR

    def get_percent_color(self, func):
        if func.addr not in self.function_info:
            self._calc_function_info(func)

        return self.function_info[func.addr]["color"]

    def get_coverage(self, func):
        if func.addr not in self.function_info:
            self._calc_function_info(func)

        return self.function_info[func.addr]["coverage"]

    def get_any_trace(self, addr):
        raise NotImplementedError

    def runtime_to_project_addr(self, addr):
        return addr - self.runtime_baddr + self.project_baddr

    def project_to_runtime_addr(self, addr):
        return addr - self.project_baddr + self.runtime_baddr

    def addr_hash(self, addr):
        return ((addr >> 4) ^ (addr << 8)) & (self.bitmap_size - 1)

    def possible_dynamic_basic_block_succs(self, g, node):
        # we return two types of edges, may_takes and fallthroughs
        may_takes = []
        fallthroughs = []

        out_edges = g.out_edges(node, data=True)
        for _, dst, data in out_edges:
            type_ = data.get("type", "transition")
            if type_ == "fake_return":
                fallthroughs.append(dst)
            else:
                may_takes.append(dst)
        return may_takes, fallthroughs

    def _incoming_transition_edges(self, g, node):
        in_edges = g.in_edges(node, data=True)
        r = []
        for src, _, data in in_edges:
            type_ = data.get("type", "transition")
            if type_ in ("transition", "exception"):
                r.append((src, node))
        return r

    def _parse_bitmap(self, func):
        func_graph = func.transition_graph
        worklist = [(func.startpoint, None)]
        done = set()
        hitcount_graph = nx.DiGraph()
        while worklist:
            node, actual_addr = worklist.pop()
            if node in done:
                continue

            hitcount_graph.add_node(node)

            may_takes, fallthroughs = self.possible_dynamic_basic_block_succs(func_graph, node)
            if len(may_takes) == 1 and not fallthroughs:
                # a continuous block might be broken into two or more because of CFG normalization, without any
                # fallthrough edges.
                # if a covered block has only one possible successor, the successor will definitely be covered
                succ = may_takes[0]
                hitcount_graph.add_node(succ)
                hitcount_graph.add_edge(node, succ, hitcount=1)  # it may not be 1 but it's hard to figure out the real
                # number
                _l.debug("%r -> %r (single successor, no fallthrough)", node, succ)
                if len(self._incoming_transition_edges(func_graph, succ)) > 1:
                    _l.debug("... %r is probably a result of graph normalization.", succ)
                    worklist.append((succ, node.addr))  # the actual address for AFL address hashing is the address of
                    # node
                else:
                    _l.debug("... %r is the target of a jump.", succ)
                    worklist.append((succ, None))

                done.add(node)
                continue

            possible_node_addrs = [node.addr]
            if actual_addr is not None:
                possible_node_addrs.append(actual_addr)

            for node_addr in possible_node_addrs:
                added = False
                for succ in may_takes:
                    prev_loc = self.addr_hash(self.project_to_runtime_addr(node_addr)) >> 1
                    cur_loc = self.addr_hash(self.project_to_runtime_addr(succ.addr))

                    idx = prev_loc ^ cur_loc
                    hitc = self.virgin_bitmap[idx]
                    _l.debug("%#x -> %#x [%#x^%#x = %#x] = %#x", node.addr, succ.addr, prev_loc, cur_loc, idx, hitc)

                    if hitc > 0:
                        added = True
                        hitcount_graph.add_node(succ)
                        hitcount_graph.add_edge(node, succ, hitcount=hitc)
                        if not isinstance(succ, Function):
                            worklist.append((succ, None))
                if added:
                    # we guessed the correct address
                    break

            if fallthroughs:
                # this may be a call, which also must be taken, but the fallthrough edges may or may not be taken
                # depending on whether the program returned from the call or not
                for fallthrough in fallthroughs:
                    _l.debug("%r -?-> %r (fallthrough)", node, fallthrough)
                    worklist.append((fallthrough, None))

            done.add(node)

        for node in hitcount_graph.nodes():
            succ_hitcount = sum(data["hitcount"] for o, data in hitcount_graph.succ[node].items())
            pred_hitcount = sum(data["hitcount"] for o, data in hitcount_graph.pred[node].items())

            node_hitc = max(succ_hitcount, pred_hitcount)
            _l.debug("Marking node %r with hitcount %d", repr(node), node_hitc)
            hitcount_graph.nodes[node]["hitcount"] = node_hitc

        return hitcount_graph

    def _calc_function_info(self, func):
        node_hitcounts = self._node_hitcounts[func]

        block_addrs = list(func.block_addrs)
        hit_count = 0

        for block_addr in block_addrs:
            if block_addr not in node_hitcounts:
                # not hit at all
                continue

            if node_hitcounts[block_addr] > 0:
                hit_count += 1

        if hit_count == 0:
            self.function_info[func.addr] = {"color": AFLQemuBitmap.FUNCTION_NOT_VISITED_COLOR, "coverage": 0}
        elif hit_count == len(block_addrs):
            self.function_info[func.addr] = {"color": AFLQemuBitmap.HIT_COLOR, "coverage": 100}
        else:
            hit_percent = (hit_count / len(block_addrs)) * 100
            bucket_size = 100 / len(AFLQemuBitmap.BUCKET_COLORS)
            bucket_pos = math.floor(hit_percent / bucket_size)
            self.function_info[func.addr] = {"color": AFLQemuBitmap.BUCKET_COLORS[bucket_pos], "coverage": hit_percent}



import atexit
import logging
from datetime import datetime
from logging.handlers import QueueHandler, QueueListener
from multiprocessing import Queue
from typing import Optional

from angr.utils.mp import Initializer

from angrmanagement.config import Conf


class LogTimeStamp:
    """
    A Log timestamp with formatting
    """

    def __init__(self, unix_timestamp: int):
        """
        :param unix_time: The unix time the timestamp represents
        """
        self._ts = datetime.fromtimestamp(unix_timestamp)
        self._cache_key: Optional[str] = None
        self._cache_str: Optional[str] = None

    def __str__(self) -> str:
        """
        Return the timestamp as a formatted string
        """
        if Conf.log_timestamp_format != self._cache_key:
            self._cache_str = self._ts.strftime(Conf.log_timestamp_format)
        return self._cache_str


class LogRecord:
    """
    Stores a log record.
    """

    __slots__ = (
        "level",
        "timestamp",
        "source",
        "content",
    )

    def __init__(self, level, unix_timestamp, source, content):
        self.timestamp = LogTimeStamp(unix_timestamp)
        self.level = level
        self.source = source
        self.content = content


class LogDumpHandler(logging.Handler):
    """
    Dumps log messages.
    """

    def __init__(self, instance, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.instance = instance

    def emit(self, record: logging.LogRecord) -> None:
        log_record = LogRecord(record.levelno, record.created, record.name, self.format(record))
        self.instance.log.append(log_record)
        self.instance.log.am_event(log_record=log_record)


class AMQueueHandler(QueueHandler):
    """
    A logging QueueHandler that is of a different type than the default QueueHandler
    This allows checking isinstance to ensure the handler is what we desired
    """


def install_queue_handler(queue: Queue):
    """
    Install a queue handler using the given queue
    This function should work for both fork and spawn modes of multiprocessing
    Fork modes may already have the parent logger installed, spawn may not
    """
    if not any(isinstance(i, AMQueueHandler) for i in logging.root.handlers):
        logging.root.handlers.insert(0, AMQueueHandler(queue))


def initialize(*args, **kwargs) -> None:
    """
    Installs a LogDumpHandler and sets up forwarding from other processes to this one
    """
    queue = Queue()
    # Install queue handlers to the current process and all future subprocesses
    Initializer.get().register(install_queue_handler, queue)
    install_queue_handler(queue)
    # Install a listener which forwards log records to the LogDumpHandler
    listener = QueueListener(queue, LogDumpHandler(*args, **kwargs))
    atexit.register(listener.stop)
    listener.start()



# pylint:disable=no-self-use

import unittest

from angrmanagement.data.tagged_interval_map import TaggedIntervalMap


class TaggedIntervalMapTests(unittest.TestCase):
    """
    Test cases for TaggedIntervalMap
    """

    def test_non_overlapping(self):
        im = TaggedIntervalMap()
        im.add(0, 100, 1)
        im.add(100, 100, 2)
        im.add(200, 100, 4)
        assert im._map == {0: 1, 100: 2, 200: 4, 300: 0}

    def test_non_overlapping_with_gaps(self):
        im = TaggedIntervalMap()
        im.add(0, 100, 1)
        im.add(200, 100, 2)
        im.add(400, 100, 4)
        assert im._map == {0: 1, 100: 0, 200: 2, 300: 0, 400: 4, 500: 0}

    def test_insertion(self):
        im = TaggedIntervalMap()
        im.add(0, 100, 1)
        im.add(25, 50, 2)
        assert im._map == {0: 1, 25: 3, 75: 1, 100: 0}

    def test_insertion_already_covered(self):
        im = TaggedIntervalMap()
        im.add(0, 100, 1)
        im.add(20, 20, 1)
        assert im._map == {0: 1, 100: 0}

    def test_overlapping(self):
        im = TaggedIntervalMap()
        im.add(0, 100, 1)
        im.add(50, 100, 2)
        im.add(100, 100, 4)
        assert im._map == {0: 1, 50: 3, 100: 6, 150: 4, 200: 0}

    def test_merge_middle(self):
        im = TaggedIntervalMap()
        im.add(0, 100, 1)
        im.add(200, 100, 1)
        im.add(100, 100, 1)
        assert im._map == {0: 1, 300: 0}

        im = TaggedIntervalMap()
        im.add(0, 10, 2)
        im.add(10, 100, 1)
        im.add(200, 100, 1)
        im.add(300, 10, 2)
        im.add(100, 100, 1)
        assert im._map == {0: 2, 10: 1, 300: 2, 310: 0}

    def test_merge_left(self):
        im = TaggedIntervalMap()
        im.add(0, 100, 1)
        im.add(100, 100, 1)
        assert im._map == {0: 1, 200: 0}
        im.add(200, 100, 1)
        assert im._map == {0: 1, 300: 0}

    def test_merge_right(self):
        im = TaggedIntervalMap()
        im.add(200, 100, 1)
        im.add(100, 100, 1)
        im.add(0, 100, 1)
        assert im._map == {0: 1, 300: 0}

    def test_binning(self):
        im = TaggedIntervalMap(4)
        im.add(2, 4, 1)
        assert im._map == {0: 1, 16: 0}
        im.add(15, 3, 2)
        assert im._map == {0: 3, 16: 2, 32: 0}

    def test_empty_iteration(self):
        im = TaggedIntervalMap()
        assert not list(im.irange())
        assert not list(im.irange(0, 100))

    def test_unbounded_iteration(self):
        im = TaggedIntervalMap()
        im.add(100, 100, 2)
        assert list(im.irange()) == [(100, 100, 2)]
        im.add(200, 100, 4)
        assert list(im.irange()) == [(100, 100, 2), (200, 100, 4)]
        im.add(0, 200, 1)
        assert list(im.irange()) == [(0, 100, 1), (100, 100, 3), (200, 100, 4)]

    def test_bounded_iteration(self):
        im = TaggedIntervalMap()
        im.add(100, 100, 1)
        im.add(150, 100, 2)
        im.add(300, 50, 4)
        assert not list(im.irange(0, 0))
        assert list(im.irange(0, 100)) == [(100, 50, 1)]
        assert list(im.irange(75, 175)) == [(100, 50, 1), (150, 50, 3)]
        assert list(im.irange(None, 175)) == [(100, 50, 1), (150, 50, 3)]
        assert list(im.irange(75, 299)) == [(100, 50, 1), (150, 50, 3), (200, 50, 2), (250, 50, 0)]
        assert list(im.irange(175, None)) == [(150, 50, 3), (200, 50, 2), (250, 50, 0), (300, 50, 4)]
        assert not list(im.irange(351, 400))


if __name__ == "__main__":
    unittest.main()



import logging

from PySide6.QtCore import QSize
from PySide6.QtWidgets import QHBoxLayout

from angrmanagement.ui.widgets.qlog_widget import QLogWidget

from .view import BaseView

_l = logging.getLogger(name=__name__)


class LogView(BaseView):
    """
    Log view displays logging output.
    """

    def __init__(self, workspace, instance, default_docking_position, *args, **kwargs):
        super().__init__("log", workspace, instance, default_docking_position, *args, **kwargs)

        self.base_caption = "Log"
        self._log_widget: QLogWidget = None

        self._init_widgets()
        self.reload()

    def closeEvent(self, event):
        self._log_widget.close()
        super().closeEvent(event)

    def reload(self):
        pass

    @staticmethod
    def minimumSizeHint(*args, **kwargs):  # pylint: disable=unused-argument
        return QSize(0, 50)

    def _init_widgets(self):
        self._log_widget = QLogWidget(self)

        hlayout = QHBoxLayout()
        hlayout.addWidget(self._log_widget)
        hlayout.setContentsMargins(0, 0, 0, 0)

        self.setLayout(hlayout)



from typing import Optional

from PySide6.QtCore import QSize
from PySide6.QtWidgets import QVBoxLayout

from angrmanagement.ui.views.view import BaseView
from angrmanagement.ui.widgets import QTraceMap


class TraceMapView(BaseView):
    """
    View container for QTraceMap.
    """

    def __init__(self, workspace, instance, default_docking_position, *args, **kwargs):
        super().__init__("tracemap", workspace, instance, default_docking_position, *args, **kwargs)
        self.base_caption: str = "Trace Map"
        self.inner_widget: Optional[QTraceMap] = None
        self._init_widgets()

    @staticmethod
    def minimumSizeHint(*args, **kwargs):  # pylint:disable=unused-argument
        return QSize(25, 25)

    @staticmethod
    def sizeHint():
        return QSize(25, 25)

    def _init_widgets(self):
        """
        Initialize widgets for this view.
        """
        self.inner_widget = QTraceMap(self.instance, parent=self)
        lyt = QVBoxLayout()
        lyt.setContentsMargins(0, 0, 0, 0)
        lyt.addWidget(self.inner_widget)
        self.setLayout(lyt)



# FIXME:
# - Add symbol resolve once https://github.com/keystone-engine/keystone/issues/351 is fixed
# - Show symbols in disassembly text
# - Support editing existing patches
# - Handle overlap with existing patches

from typing import TYPE_CHECKING, Optional

from PySide6.QtCore import QSize
from PySide6.QtGui import QFont
from PySide6.QtWidgets import (
    QCheckBox,
    QDialog,
    QDialogButtonBox,
    QGridLayout,
    QLabel,
    QLineEdit,
    QVBoxLayout,
)

try:
    import keystone
except ImportError:
    keystone = None

from angr.knowledge_plugins.patches import Patch
from archinfo import ArchError

from angrmanagement.config import Conf

if TYPE_CHECKING:
    from angrmanagement.data.instance import Instance


class AssemblePatchDialog(QDialog):
    """
    Dialog for making a patch from assembly code.
    """

    def __init__(self, address: int, instance: "Instance", parent=None):
        super().__init__(parent)

        self.instance: Instance = instance
        self._patch_addr: int = address

        block = self.instance.project.factory.block(self._patch_addr)
        insn = block.disassembly.insns[0]

        self._original_bytes: bytes = block.bytes[: insn.size]
        self._new_bytes: Optional[bytes] = self._original_bytes
        self._initial_text = insn.mnemonic
        if insn.op_str:
            self._initial_text += " " + insn.op_str

        self._init_widgets()
        self.setWindowTitle(f"Assemble Patch at {self._patch_addr:#x}")

    def sizeHint(self):  # pylint:disable=no-self-use
        return QSize(400, 150)

    #
    # Private methods
    #

    def _init_widgets(self):
        self.main_layout = QVBoxLayout()
        font = QFont(Conf.disasm_font)

        grid_layout = QGridLayout()
        grid_layout.addWidget(QLabel("Assembly:", self), 0, 0)
        self._insn_text = QLineEdit(self)
        self._insn_text.setFont(font)
        self._insn_text.setText(self._initial_text)
        self._insn_text.selectAll()
        grid_layout.addWidget(self._insn_text, 0, 1)

        grid_layout.addWidget(QLabel("Bytes:", self), 1, 0)
        self._bytes_text = QLineEdit(self)
        self._bytes_text.setFont(font)
        self._bytes_text.setReadOnly(True)
        grid_layout.addWidget(self._bytes_text, 1, 1)

        self._pad_checkbox = QCheckBox("Pad to original instruction size", self)
        self._pad_checkbox.setChecked(True)
        grid_layout.addWidget(self._pad_checkbox, 2, 1)

        self.main_layout.addLayout(grid_layout)
        self.main_layout.addStretch(1)
        self._status_label = QLabel(self)
        self.main_layout.addWidget(self._status_label)

        buttons = QDialogButtonBox(parent=self)
        buttons.setStandardButtons(QDialogButtonBox.StandardButton.Cancel | QDialogButtonBox.StandardButton.Ok)
        buttons.accepted.connect(self._on_ok_clicked)
        buttons.rejected.connect(self.close)
        self._ok_button = buttons.button(QDialogButtonBox.Ok)
        self._ok_button.setEnabled(False)
        self.main_layout.addWidget(buttons)

        self.setLayout(self.main_layout)

        can_assemble = False
        status_msg = ""
        if keystone:
            try:
                ks = self.instance.project.arch.keystone
                can_assemble = ks is not None
            except ArchError as e:
                status_msg = str(e)
        else:
            status_msg = "Keystone not installed"

        if can_assemble:
            self._assemble()
            self._insn_text.textChanged.connect(self._on_text_changed)
            self._pad_checkbox.stateChanged.connect(self._on_checkbox_changed)
        else:
            self._insn_text.setEnabled(False)
            self._bytes_text.setEnabled(False)
            self._pad_checkbox.setEnabled(False)
            self._update_status(status_msg, False)

    def _assemble(self):
        success = False
        status_msg = ""

        try:
            ks = self.instance.project.arch.keystone

            text = self._insn_text.text()
            if len(text) > 0:
                self._new_bytes = ks.asm(self._insn_text.text(), self._patch_addr, as_bytes=True)[0] or b""
            else:
                self._new_bytes = b""

            # Pad to original instruction length
            byte_length_delta = len(self._original_bytes) - len(self._new_bytes)
            if byte_length_delta > 0:
                if self._pad_checkbox.isChecked():
                    nop_instruction_bytes = self.instance.project.arch.nop_instruction
                    self._new_bytes += (byte_length_delta // len(nop_instruction_bytes)) * nop_instruction_bytes
                    byte_length_delta = len(self._original_bytes) - len(self._new_bytes)
                    if byte_length_delta:
                        status_msg = "Warning: Unable to completely pad remainder"
            elif byte_length_delta < 0:
                status_msg = "Warning: Patch exceeds original instruction length"

            success = True
        except keystone.KsError as ks_error:
            self._new_bytes = b""
            status_msg = "Error: " + ks_error.message

        self._update_status(status_msg, success and len(self._new_bytes) > 0)

    def _update_status(self, status_msg, can_press_ok):
        self._bytes_text.setText(self._new_bytes.hex(" "))
        self._status_label.setText(status_msg)
        self._status_label.setProperty("class", "status_invalid")
        self._status_label.style().unpolish(self._status_label)
        self._status_label.style().polish(self._status_label)
        self._ok_button.setEnabled(can_press_ok)

    #
    # Event handlers
    #

    def _on_text_changed(self, new_text):  # pylint: disable=unused-argument
        self._assemble()

    def _on_checkbox_changed(self, state):  # pylint: disable=unused-argument
        self._assemble()

    def _on_ok_clicked(self):
        if self._new_bytes != self._original_bytes:
            pm = self.instance.project.kb.patches

            # XXX: Currently patch manager stores patches by address, so remove any existing patch at this addr
            existing_patch = pm.get_patch(self._patch_addr)
            if existing_patch:
                self.instance.patches.remove_patch(existing_patch.addr)
                self.instance.patches.am_event(removed={existing_patch})

            patch = Patch(self._patch_addr, self._new_bytes)
            pm.add_patch_obj(patch)
            self.instance.patches.am_event(added={patch})

        self.accept()



import logging

from PySide6.QtCore import QRectF, Qt
from PySide6.QtGui import QColor, QPen
from PySide6.QtWidgets import QGraphicsItem

from angrmanagement.config import Conf
from angrmanagement.utils import locate_function

_l = logging.getLogger(__name__)


class QStateBlock(QGraphicsItem):
    HORIZONTAL_PADDING = 5
    VERTICAL_PADDING = 5
    LINE_MARGIN = 3

    def __init__(self, is_selected, symexec_view, state=None, history=None):
        super().__init__()

        self.symexec_view = symexec_view
        self._instance = self.symexec_view.instance
        self._config = Conf

        self.state = state
        self.history = history
        if history is None and state is not None:
            self.history = state.history
        if history is not None and state is None:
            self.state = history.state
        self.selected = is_selected

        # widgets
        self._label_str = None
        self._function_str = None

        self.addr = None

        self._init_widgets()
        self._update_size()

    def get_state(self):
        if self.state is not None:
            return self.state
        elif self.history is not None:
            return self.history.state
        else:
            return None

    def _init_widgets(self):
        addr = None
        if self.state.regs._ip.symbolic:
            self._label_str = str(self.state.regs._ip)
        else:
            addr = self.state.regs._ip._model_concrete.value
            self._label_str = "%#x" % addr
        self._label_str = "State " + self._label_str

        self.addr = addr

        if addr is None:
            self._function_str = "Unknown"
        else:
            the_func = locate_function(self._instance, addr)
            if the_func is None:
                # is it a SimProcedure?
                if self._instance.project.is_hooked(addr):
                    hooker = self._instance.project.hooked_by(addr)
                    self._function_str = "SimProcedure " + hooker.__class__.__name__.split(".")[-1]
                else:
                    self._function_str = "Unknown"
            else:
                offset = addr - the_func.addr
                if not the_func.name:
                    self._function_str = f"{the_func.addr:#x}{offset:+x}"
                else:
                    self._function_str = f"{the_func.name}{offset:+x}"
        self._function_str = "Function: %s" % self._function_str

    def mousePressEvent(self, event):  # pylint: disable=no-self-use
        super().mousePressEvent(event)

    def mouseReleaseEvent(self, event):
        # _l.debug('QStateBlock received mouse release event')
        if event.button() == Qt.LeftButton:
            self.selected = not self.selected
            self.symexec_view.redraw_graph()
            event.accept()

        super().mouseReleaseEvent(event)

    def mouseDoubleClickEvent(self, event):
        # _l.debug('QStateBlock received mouse double click event')
        if event.button() == Qt.LeftButton:
            if self.state is not None:
                self.symexec_view.workspace.viz(self.state.addr)
                event.accept()
            elif self.history is not None:
                self.symexec_view_instance.workspace.viz(self.history.state.addr)
                event.accept()

        super().mouseDoubleClickEvent(event)

    def paint(self, painter, option, widget):  # pylint: disable=unused-argument
        """
        Paint a state block on the scene.

        :param painter:
        :return: None
        """

        painter.setFont(Conf.symexec_font)
        normal_background = QColor(0xFA, 0xFA, 0xFA)
        selected_background = QColor(0xCC, 0xCC, 0xCC)

        # The node background
        if self.selected:
            painter.setBrush(selected_background)
        else:
            painter.setBrush(normal_background)
        painter.setPen(QPen(QColor(0xF0, 0xF0, 0xF0), 1.5))
        painter.drawRect(0, 0, self.width, self.height)

        x = 0
        y = 0

        # The addr label
        addr_label_x = x + self.HORIZONTAL_PADDING
        addr_label_y = y + self.VERTICAL_PADDING
        painter.setPen(Qt.black)
        painter.drawText(addr_label_x, addr_label_y + self._config.symexec_font_ascent, self._label_str)

        y += self._config.symexec_font_height + self.LINE_MARGIN
        x = 0

        # The function label
        function_label_x = x + self.HORIZONTAL_PADDING
        function_label_y = y + self.VERTICAL_PADDING
        painter.drawText(function_label_x, function_label_y + self._config.symexec_font_ascent, self._function_str)

    @property
    def height(self):
        return self.boundingRect().height()

    @property
    def width(self):
        return self.boundingRect().width()

    def boundingRect(self):
        return QRectF(0, 0, self._width, self._height)

    #
    # Private methods
    #

    def _update_size(self):
        width_candidates = [
            self.HORIZONTAL_PADDING * 2 + len(self._label_str) * self._config.symexec_font_width,
            self.HORIZONTAL_PADDING * 2 + len(self._function_str) * self._config.symexec_font_width,
        ]
        height_candidates = [0]
        self._width = max(width_candidates)
        self._height = max(height_candidates)

        self._width = max(100, self._width)
        self._height = max(50, self._height)



from typing import Iterator, Optional, Tuple

from sortedcontainers import SortedDict


class TaggedIntervalMap:
    """
    Catalogs features of intervals.
    """

    def __init__(self, nbits: int = 0):
        """
        :param nbits: Number of binning bits. Higher values reduce detail. 0 for no binning.
        """
        self._nbits: int = nbits
        self._map: SortedDict = SortedDict()  # SortedDict[int, int]

    @property
    def nbits(self) -> int:
        return self._nbits

    def add(self, addr: int, size: int, tags: int) -> None:
        """
        Add interval starting at `addr` of `size` bytes.

        When binning, intervals endpoints are aligned to 2^nbits. Gaps between added intervals are populated with
        implicit intervals having tag value of 0. Overlapping intervals will have tag bits OR'd together.

        Adjacent intervals in the interval map have unique tags. When intervals are added to the map, any adjacent stops
        with identical tags will be eliminated so the map retains this property.

        For example: if an interval(addr=0, size=100, tags=1) is added, followed by (100, 100, 1), the resulting
        interval in the map will be (0, 200, 1).
        """
        assert addr >= 0
        assert size >= 0
        assert tags != 0

        if size == 0:
            return

        max_bin_offset = (1 << self._nbits) - 1
        mask = ~max_bin_offset

        start_addr = addr & mask  # Round down to bin alignment
        end_addr = (addr + size + max_bin_offset) & mask  # Round up to bin alignment

        if self._is_already_covered(start_addr, end_addr, tags):
            return

        self._insert_stop(start_addr)
        self._insert_stop(end_addr)
        for affected_addr in self._map.irange(start_addr, end_addr, inclusive=(True, False)):
            self._map[affected_addr] |= tags
        self._eliminate_extraneous_stops(start_addr, end_addr)

    def _insert_stop(self, addr: int) -> None:
        """
        Insert a new interval stop point at `addr`, if one is not already present in the map. Tags are copied from
        nearest stop before `addr`.
        """
        if addr not in self._map:
            idx = self._map.bisect(addr) - 1
            self._map[addr] = self._map.peekitem(idx)[1] if idx >= 0 else 0

    def _is_already_covered(self, min_addr: int, end_addr: int, tags: int) -> bool:
        """
        Determine if interval [min_addr, end_addr) is covered by an existing range with identical tags.
        """
        idx = self._map.bisect(min_addr) - 1
        if idx >= 0 and len(self._map) > idx + 1:
            e_addr, e_tags = self._map.peekitem(idx)
            e_addr_next, _ = self._map.peekitem(idx + 1)
            return (e_addr <= min_addr) and (end_addr <= e_addr_next) and (e_tags == tags)
        return False

    def _eliminate_extraneous_stops(self, min_addr: int, max_addr: int) -> None:
        """
        Canonicalize the map by eliminating adjacent stops with identical tags both inside and directly outside of
        [min_addr, max_addr].
        """
        keys_to_drop = []
        prev_tags = None
        for addr, _, tags in self.irange(min_addr, max_addr):
            if tags == prev_tags:
                keys_to_drop.append(addr)
            else:
                prev_tags = tags

        for addr in keys_to_drop:
            del self._map[addr]

    def irange(self, min_addr: Optional[int] = None, max_addr: Optional[int] = None) -> Iterator[Tuple[int, int, int]]:
        """
        Iterate over intervals intersecting [min_addr, max_addr], yielding interval (addr, size, tags) tuples. Implicit
        gap intervals (with tags=0) are also returned.

        :param min_addr: Minimum address (inclusive) to begin iterating from. If None, iterate from start of map.
        :param max_addr: Maximum address (inclusive) to iterate to. If None, iterate to end of map.
        """
        if not self._map:
            return

        start_idx = 0 if min_addr is None else max(0, self._map.bisect_left(min_addr) - 1)
        stop_idx = None if max_addr is None else (self._map.bisect(max_addr) + 1)

        start_addr, tags = self._map.peekitem(start_idx)
        for addr in self._map.islice(start_idx + 1, stop_idx):
            yield (start_addr, addr - start_addr, tags)
            tags = self._map[addr]
            start_addr = addr



import threading

import archr
from PySide6.QtCore import Qt
from PySide6.QtWidgets import QHBoxLayout, QMainWindow, QMessageBox, QPushButton, QVBoxLayout
from qtterm import TerminalWidget

from angrmanagement.plugins import BasePlugin
from angrmanagement.ui.views import BaseView
from angrmanagement.ui.views.interaction_view import (
    PlainTextProtocol,
    SavedInteraction,
)


class ConsoleView(BaseView):
    """
    ConsoleView
    """

    def __init__(self, workspace, *args, **kwargs):
        super().__init__("interaction console", workspace, *args, **kwargs)

        self.base_caption = "Interaction Console"
        self.workspace = workspace
        self.target = None
        self.conversations = {}
        self.terminal = TerminalWidget(command=None)
        self.analyzer = None
        self.interaction_context = None

        main_layout = QVBoxLayout()
        controls_layout = QHBoxLayout()

        connect_button = QPushButton()
        connect_button.setText("Connect")
        connect_button.clicked.connect(self.connect)
        controls_layout.addWidget(connect_button)

        terminal_window = QMainWindow()
        terminal_window.setWindowFlags(Qt.Widget)
        terminal_window.setCentralWidget(self.terminal)

        main_layout.addLayout(controls_layout)
        main_layout.addWidget(terminal_window)

        self.setLayout(main_layout)

    def connect(self):
        self.disconnect()

        img_name = self.workspace.main_instance.img_name
        if img_name is None:
            QMessageBox.critical(None, "Nothing to run", "The project was not loaded from a docker image")
            return

        self.target = archr.targets.DockerImageTarget(img_name, companion=True)
        self.target.build().start()

        self.analyzer = archr.analyzers.TCPDumpAnalyzer(self.target)

        self.interaction_context = self.analyzer.fire_context(timeout_exception=False)
        self.interaction_context.__enter__()  # pylint:disable=no-member

        threading.Thread(target=self._inotify_thread, daemon=True).start()

        self.workspace.extract_conversations = self.analyzer.extract_conversations

        self.terminal.execute(
            [
                "docker",
                "exec",
                "-it",
                self.target.companion_container.id,
                "bash",
            ]
        )

    def disconnect(self):
        if self.target:
            self.terminal.stop()
            self.target.__exit__()

    def _inotify_thread(self):
        inotify = self.target.run_companion_command(
            [
                "inotifywait",
                "--monitor",
                "--event",
                "modify",
                archr.analyzers.TCPDumpAnalyzer.pcap_path,
            ]
        )

        response = inotify.stderr.read(0x1000)
        if b"Watches established." not in response:
            raise Exception("Failed to setup TCPDump watcher!")

        while True:
            inotify.stdout.read(0x1000)
            conversations = self.analyzer.extract_conversations()
            for id_, conversation in conversations.items():
                if id_ not in self.conversations:
                    self._save_interaction(conversation)

            self.conversations = conversations

    def _save_interaction(self, conversation):
        target_port = self.target.tcp_ports[0]

        def direction(srcport, dstport):
            if srcport == target_port:
                return "out"
            if dstport == target_port:
                return "in"
            raise Exception("Unknown direction")

        log = [
            {
                "dir": direction(srcport, dstport),
                "data": payload,
            }
            for (srcport, dstport, payload) in conversation
        ]

        name = hex(hash(str(conversation)) & 0xFFFFFFFFFFFFFFFF)[2:].rjust(16, "0")

        self.workspace.main_instance.interactions.am_obj.append(SavedInteraction(name, PlainTextProtocol, log))
        self.workspace.main_instance.interactions.am_event()


class InteractionConsole(BasePlugin):
    """
    InteractionConsole Plugin
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.console_view = ConsoleView(self.workspace, "center")
        self.workspace.default_tabs += [self.console_view]
        self.workspace.add_view(self.console_view)

    def teardown(self):
        if self.console_view.target:
            self.console_view.target.__exit__()



from PySide6.QtCore import QSize, Qt
from PySide6.QtWidgets import QFrame, QHBoxLayout, QLabel, QScrollArea, QSizePolicy, QVBoxLayout

from .qast_viewer import QASTViewer


class QVEXTempsViewer(QFrame):
    def __init__(self, state, parent, workspace):
        super().__init__(parent)
        self.workspace = workspace

        self.state = state

        # widgets
        self._area = None
        self._tmps = {}

        self._init_widgets()

    #
    # Overridden methods
    #

    def sizeHint(self, *args, **kwargs):
        return QSize(100, 100)

    #
    # Private methods
    #

    def _init_widgets(self):
        area = QScrollArea()
        area.setVerticalScrollBarPolicy(Qt.ScrollBarAsNeeded)
        area.setHorizontalScrollBarPolicy(Qt.ScrollBarAsNeeded)
        area.setWidgetResizable(True)

        self._area = area

        base_layout = QVBoxLayout()
        base_layout.addWidget(area)
        self.setLayout(base_layout)

    def _load_tmps(self):
        state = self.state.am_obj

        layout = QVBoxLayout()

        self._tmps.clear()
        tmps = {} if state is None else state.scratch.temps

        # tmps
        for tmp_id, tmp_value in tmps.items():
            sublayout = QHBoxLayout()

            lbl_tmp_name = QLabel(self)
            lbl_tmp_name.setProperty("class", "reg_viewer_label")
            lbl_tmp_name.setText("tmp_%d" % tmp_id)
            lbl_tmp_name.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
            sublayout.addWidget(lbl_tmp_name)

            sublayout.addSpacing(10)

            tmp_viewer = QASTViewer(tmp_value, workspace=self.workspace, parent=self)
            self._tmps[tmp_id] = tmp_viewer
            sublayout.addWidget(tmp_viewer)

            layout.addLayout(sublayout)

        layout.setSpacing(0)
        layout.addStretch(0)
        layout.setContentsMargins(2, 2, 2, 2)

        # the container
        container = QFrame()
        container.setAutoFillBackground(True)
        palette = container.palette()
        palette.setColor(container.backgroundRole(), Qt.white)
        container.setPalette(palette)
        container.setLayout(layout)

        self._area.setWidget(container)

    def _watch_state(self, **kwargs):
        self._load_tmps()



import importlib
import logging
import os
import sys
from typing import List, Tuple

from .base_plugin import BasePlugin
from .plugin_description import PluginDescription

log = logging.getLogger(__name__)


def load_plugin_descriptions_from_dir(path: str) -> List[Tuple[str, PluginDescription]]:
    try:
        dlist = os.listdir(path)
    except OSError:
        return []

    plugins = []
    for d in dlist:
        descs = load_plugin_description(os.path.join(path, d))
        for desc in descs:
            plugins.append((d, desc))
    return plugins


def load_plugin_description(path: str) -> List[PluginDescription]:
    try:
        flist = os.listdir(path)
    except OSError:
        return []

    if "plugin.toml" in flist:
        fpath = os.path.join(path, "plugin.toml")
        if os.path.isfile(fpath):
            return PluginDescription.from_toml(fpath)
    return []


def load_default_plugins():
    return load_plugins_from_dir(
        os.path.dirname(os.path.abspath(__file__)),
        exclude=("base_plugin.py", "plugin_manager.py", "plugin_description.py"),
    )


def load_plugins_from_dir(path, exclude=()):
    # hack.
    if not exclude and path == os.path.dirname(__file__):
        exclude = ("base_plugin.py", "plugin_manager.py")

    out = []
    try:
        dlist = os.listdir(path)
    except OSError:
        return []

    for filename in dlist:
        if filename in exclude or filename in ("__init__.py", "__pycache__"):
            continue
        fullname = os.path.join(path, filename)
        if os.path.isfile(fullname) and fullname.endswith(".py"):
            out += load_plugins_from_file(fullname)
        elif os.path.isfile(os.path.join(fullname, "__init__.py")):
            out += load_plugins_from_package(fullname)

    return out


def load_plugins_from_package(path):
    # this logic is a little multiplexed in the next function but like... whatever
    return load_plugins_from_file(os.path.join(path, "__init__.py"))


def load_plugins_from_file(path):
    basename = os.path.basename(path)
    if basename == "__init__.py":
        modbasename = os.path.basename(os.path.dirname(path))
        if modbasename.count(".") != 0:
            log.error("file %s cannot be loaded - weird name", path)
            return []
    else:
        if os.path.isfile(path):
            modbasename = os.path.basename(os.path.dirname(path))
            modbasename += "." + basename.split(".")[0]
            if modbasename.count(".") != 1:
                log.error("package %s cannot be loaded - weird name", path)
                return []
        else:
            # directory
            modbasename = basename.split(".")[0]
            path = os.path.join(path, "__init__.py")
            if basename.count(".") != 0:
                log.error("package %s cannot be loaded - weird name", path)
                return []
    modname = "angrmanagement.plugins.%s" % modbasename

    # https://stackoverflow.com/questions/67631/how-to-import-a-module-given-the-full-path
    spec = importlib.util.spec_from_file_location(modname, path, submodule_search_locations=[])
    if spec is None:
        log.error("Not a python module: %s", path)
        return []
    try:
        mod = importlib.util.module_from_spec(spec)
        sys.modules[modname] = mod
        spec.loader.exec_module(mod)
    except Exception as e:
        return [e]

    return load_plugins_from_module(mod)


def load_plugins_from_module(module):
    return load_plugins_from_vars(vars(module))


def load_plugins_from_vars(variables):
    out = []
    for _, cls in variables.items():
        if (
            type(cls) is type
            and issubclass(cls, BasePlugin)
            and not hasattr(cls, "_%s__i_hold_this_abstraction_token" % cls.__name__)
        ):
            out.append(cls)
    return out



import os
import pathlib
import subprocess
import sys

from PySide6.QtCore import QSettings

from angrmanagement.utils.env import app_path


class AngrUrlScheme:
    URL_SCHEME = "angr"
    WIN_REG_PATH = "HKEY_CURRENT_USER\\Software\\Classes\\{}"

    """
    Registers and handlers URL schemes to the operating system.
    """

    def register_url_scheme(self):
        if sys.platform.startswith("win"):
            self._register_url_scheme_windows()
        elif sys.platform.startswith("linux"):
            self._register_url_scheme_linux()
        else:
            raise NotImplementedError("We currently do not support registering angr URL scheme on %s." % sys.platform)

    def unregister_url_scheme(self):
        if sys.platform.startswith("win"):
            self._unregister_url_scheme_windows()
        elif sys.platform.startswith("linux"):
            self._unregister_url_scheme_linux()
        else:
            raise NotImplementedError("We currently do not support unregistering angr URL scheme on %s." % sys.platform)

    def is_url_scheme_registered(self):
        if sys.platform.startswith("win"):
            return self._is_url_scheme_registered_windows()
        elif sys.platform.startswith("linux"):
            return self._is_url_scheme_registered_linux()
        else:
            return False, None

    def is_url_scheme_supported(self):
        return sys.platform.startswith("win") or sys.platform.startswith("linux")

    #
    # Utils
    #

    @staticmethod
    def _angr_desktop_path():
        home_dir = os.path.expanduser("~")
        p = os.path.join(home_dir, ".local", "share", "applications", "angr.desktop")
        return p

    #
    # Windows
    #

    def _register_url_scheme_windows(self):
        app_path_ = app_path(pythonw=True)

        reg_path = self.WIN_REG_PATH.format(self.URL_SCHEME)
        reg = QSettings(reg_path, QSettings.NativeFormat)

        reg.setValue("Default", "angr management")
        reg.setValue("URL Protocol", "")

        # reg.beginGroup("DefaultIcon")
        # reg.setValue("Default", TODO)
        # reg.endGroup()

        reg.beginGroup("shell")
        reg.beginGroup("open")
        reg.beginGroup("command")
        reg.setValue("Default", app_path_ + ' -u "%1"')
        reg.endGroup()
        reg.endGroup()
        reg.endGroup()

    def _unregister_url_scheme_windows(self):
        reg_path = self.WIN_REG_PATH.format(self.URL_SCHEME)
        reg = QSettings(reg_path, QSettings.NativeFormat)

        reg.remove("")

    def _is_url_scheme_registered_windows(self):
        reg_path = self.WIN_REG_PATH.format(self.URL_SCHEME)
        reg = QSettings(reg_path, QSettings.NativeFormat)

        if reg.contains("Default"):
            reg.beginGroup("shell")
            reg.beginGroup("open")
            reg.beginGroup("command")
            if reg.contains("Default"):
                return True, reg.value("Default")
        return False, None

    #
    # Linux
    #

    def _register_url_scheme_linux(self):
        cmd_0 = ["xdg-mime", "default", "angr.desktop", f"x-scheme-handler/{self.URL_SCHEME}"]

        # test if xdg-mime is available
        retcode = subprocess.call(["xdg-mime"], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        if retcode != 1:
            raise FileNotFoundError("xdg-mime is not installed.")
        retcode = subprocess.call(["xdg-mime", "--help"], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        if retcode != 0:
            raise FileNotFoundError("xdg-mime is not installed.")

        # extract angr.desktop
        angr_desktop = """[Desktop Entry]
Comment=angr management
Exec={app_path} -u %U
Hidden=true
Name=angr management
Terminal=false
MimeType=x-scheme-handler/{url_scheme};
Type=Application
"""
        angr_desktop_path = self._angr_desktop_path()
        angr_desktop_base = os.path.dirname(angr_desktop_path)
        pathlib.Path(angr_desktop_base).mkdir(parents=True, exist_ok=True)
        with open(angr_desktop_path, "w") as f:
            f.write(angr_desktop.format(app_path=app_path(), url_scheme=self.URL_SCHEME))

        # register the scheme
        retcode = subprocess.call(cmd_0)
        if retcode != 0:
            raise ValueError('Failed to setup the URL scheme. Command "%s" failed.' % " ".join(cmd_0))

    def _unregister_url_scheme_linux(self):
        angr_desktop_path = self._angr_desktop_path()
        if os.path.isfile(angr_desktop_path):
            os.unlink(angr_desktop_path)

    def _is_url_scheme_registered_linux(self):
        # angr.desktop
        angr_desktop_path = self._angr_desktop_path()
        if not os.path.isfile(angr_desktop_path):
            return False, None

        # is xdg-mime available
        retcode = subprocess.call(["xdg-mime"], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        if retcode != 1:
            return False, None

        # xdg-mine query
        proc = subprocess.Popen(
            ["xdg-mime", "query", "default", f"x-scheme-handler/{self.URL_SCHEME}"],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        stdout, _ = proc.communicate()
        if not stdout:
            return False, None

        # Load Exec=
        with open(angr_desktop_path) as f:
            data = f.read()
        lines = data.split("\n")
        cmdline = None
        for line in lines:
            if line.startswith("Exec="):
                cmdline = line[5:]
                break
        if cmdline is None:
            return False, None
        return True, cmdline



import logging
from pathlib import Path
from typing import Optional

import angr
from PySide6.QtGui import QColor
from PySide6.QtWidgets import QFileDialog

from angrmanagement.data.instance import Instance
from angrmanagement.plugins import BasePlugin
from angrmanagement.ui.views import DisassemblyView

from .diff_view import DiffDisassemblyView
from .function_diff import BFSFunctionDiff, FunctionDiff
from .settings_dialog import SettingsDialog

logger = logging.getLogger(__name__)


class PreciseDiffPlugin(BasePlugin):
    """
    A plugin for performing precisions diffs on two binaries loaded in angr-management.
    """

    DIFF_SETTINGS_CMD_EVT = 0
    LOAD_BINARY_CMD_EVT = 1
    RELOAD_BINARY_CMD_EVT = 2
    MENU_BUTTONS = ("Precise Diffing settings...", "Load binary for Precise Diffing...", "Refresh loaded diff binary")

    def __init__(self, workspace):
        super().__init__(workspace)
        self.diff_instance: Optional[Instance] = None
        self.current_revised_view: Optional[DisassemblyView] = None

        self.loaded_binary: Optional[Path] = None
        self.diff_algo: Optional[FunctionDiff] = None

        self.prefer_symbols = True
        self.resolve_strings = True
        self.diff_algo_class = BFSFunctionDiff
        self.add_color = QColor(0xDDFFDD)
        self.del_color = QColor(0xFF7F7F)
        self.chg_color = QColor(0xF4ECC2)

        self.seen_insns = set()

    #
    # UI Callback Handlers
    #

    def handle_click_menu(self, idx):
        if idx == self.DIFF_SETTINGS_CMD_EVT:
            dialog = SettingsDialog(self)
            dialog.exec()
            if dialog.updates and self.loaded_binary:
                self.syncronize_with_original_disassembly_view()

        elif idx == self.LOAD_BINARY_CMD_EVT:
            filepath, _ = QFileDialog.getOpenFileName(caption="Load Recompiled Object")
            if not filepath:
                logger.warning("Binary selection cancelled.")
                return

            self.load_revised_binary_from_file(Path(filepath))

        elif idx == self.RELOAD_BINARY_CMD_EVT and self.loaded_binary is not None:
            self.load_revised_binary_from_file(self.loaded_binary)

    def color_insn(self, addr, selected, disasm_view):
        if disasm_view != self.current_revised_view:
            return None

        if self.diff_algo is None:
            return None

        diff_value = self.diff_algo.addr_diff_value(addr)
        return self._color_map(diff_value)

    #
    # View Construction
    #

    def _color_map(self, diff_value):
        diff_map = {
            FunctionDiff.OBJ_ADDED: self.add_color,
            FunctionDiff.OBJ_DELETED: self.del_color,
            FunctionDiff.OBJ_CHANGED: self.chg_color,
            FunctionDiff.OBJ_UNMODIFIED: None,
        }

        return diff_map[diff_value]

    def color_graph_diff(self, og_disasm: DisassemblyView, new_disasm: DisassemblyView):
        try:
            base_func = og_disasm.function.am_obj
            rev_func = new_disasm.function.am_obj
        except (AttributeError, ValueError):
            return

        if base_func is None or rev_func is None:
            return

        self.diff_algo = self.diff_algo_class(
            base_func,
            rev_func,
            disas_base=og_disasm.disasm,
            disas_rev=new_disasm.disasm,
            view_base=og_disasm,
            view_rev=new_disasm,
            resolve_strings=self.resolve_strings,
            prefer_symbols=self.prefer_symbols,
        )
        new_disasm.redraw_current_graph()

    def _destroy_recompiled_view(self):
        if self.current_revised_view:
            self.workspace.remove_view(self.current_revised_view)
            del self.current_revised_view

        if self.diff_instance:
            del self.diff_instance

    def _create_instance_from_binary(self, file_path: Path):
        recompilation_instance = Instance()
        recompilation_instance.recompilation_plugin = self
        recompilation_instance.workspace = self.workspace
        recompilation_instance.project.am_obj = angr.Project(file_path, auto_load_libs=False)
        self.loaded_binary = file_path
        recompilation_instance.cfg = recompilation_instance.project.analyses.CFG()

        return recompilation_instance

    def _create_revised_disassembly_view(self):
        new_disass = DiffDisassemblyView(self.workspace, self.diff_instance, "center")
        new_disass.category = "diff"
        new_disass.base_caption = "Precise Diff"
        self.current_revised_view = new_disass
        self.workspace.add_view(self.current_revised_view)
        return self.current_revised_view

    def jump_to_in_revised_view(self, func):
        self.current_revised_view.display_function(func)
        self.current_revised_view.jump_to(func.addr)

    # pylint:disable=unused-argument
    def syncronize_with_original_disassembly_view(self, *args, **kwargs):
        og_view = self.workspace._get_or_create_view("disassembly", DisassemblyView)
        if not og_view:
            return

        try:
            func_obj = og_view.current_function
        except NotImplementedError:
            return

        if func_obj is None or func_obj.am_obj is None:
            return

        og_func = func_obj.am_obj
        og_func_name = og_func.name

        try:
            revised_func = self.diff_instance.cfg.functions[og_func_name]
        except KeyError:
            logger.warning("The function %s does not exist in the diffed binary", og_func)
            return

        self.jump_to_in_revised_view(revised_func)
        self.color_graph_diff(og_view, self.current_revised_view)

    def load_revised_binary_from_file(self, file_path: Path):
        self._destroy_recompiled_view()
        self.diff_instance = self._create_instance_from_binary(file_path)
        self._create_revised_disassembly_view()
        self.syncronize_with_original_disassembly_view()
        self.workspace.view_manager.raise_view(self.current_revised_view)

        original_disass_view = self.diff_instance.workspace._get_or_create_view("disassembly", DisassemblyView)
        self.current_revised_view.sync_with_state_object(original_disass_view.sync_state)



import logging
import os
from typing import TYPE_CHECKING, Optional, Sequence, Tuple

from angrmanagement.data.breakpoint import BreakpointType
from angrmanagement.data.trace import BintraceTrace

from .debugger import Debugger

if TYPE_CHECKING:
    from angr import SimState
    from angr.knowledge_plugins import Function

    from angrmanagement.ui.workspace import Workspace


try:
    import bintrace
    from bintrace import TraceEvent
    from bintrace.debugger_angr import AngrTraceDebugger
except ImportError:
    bintrace = None
    TraceEvent = "TraceEvent"

_l = logging.getLogger(name=__name__)


class BintraceDebugger(Debugger):
    """
    Trace playback debugger.
    """

    def __init__(self, trace: BintraceTrace, workspace: "Workspace"):
        super().__init__(workspace)
        assert bintrace is not None
        assert isinstance(trace, BintraceTrace)
        self._trace: BintraceTrace = trace
        self._btrace: bintrace.Trace = trace.trace
        self._trace_dbg: AngrTraceDebugger = AngrTraceDebugger(
            self._btrace, self.workspace.main_instance.project.am_obj
        )
        self._cached_simstate = None

    def __str__(self):
        pc = self.simstate.solver.eval(self.simstate.regs.pc)
        return f"{os.path.basename(self._btrace.path)} @ {pc:x}"

    def _on_state_change(self):
        """
        Common handler for state changes.
        """
        self._cached_simstate = None
        self.state_changed.am_event()

    def _sync_breakpoints(self):
        """
        Synchronize breakpoints set in Workspace with trace debugger.
        """
        bp_type_map = {
            BreakpointType.Execute: bintrace.debugger.BreakpointType.Execute,
            BreakpointType.Read: bintrace.debugger.BreakpointType.Read,
            BreakpointType.Write: bintrace.debugger.BreakpointType.Write,
        }
        self._trace_dbg.breakpoints = {
            bintrace.debugger.Breakpoint(bp_type_map[bp.type], bp.addr, bp.size)
            for bp in self.workspace.main_instance.breakpoint_mgr.breakpoints
        }

    @property
    def simstate(self) -> "SimState":
        if self._cached_simstate is None:
            self._cached_simstate = self._trace_dbg.simstate
        return self._cached_simstate

    @property
    def is_running(self) -> bool:
        return True

    @property
    def can_step_backward(self) -> bool:
        return self._trace_dbg.can_step_backward

    def step_backward(self):
        if self.can_step_backward:
            self._trace_dbg.step_backward()
            self._on_state_change()

    @property
    def can_step_forward(self) -> bool:
        return self._trace_dbg.can_step_forward

    def step_forward(self, until_addr: Optional[int] = None):
        if self.can_step_forward:
            self._trace_dbg.step_forward(until_addr=until_addr)
            self._on_state_change()

    @property
    def can_continue_backward(self) -> bool:
        return self._trace_dbg.can_continue_backward

    def continue_backward(self):
        if self.can_continue_backward:
            self._sync_breakpoints()
            self._trace_dbg.continue_backward()
            self._on_state_change()

    @property
    def can_continue_forward(self) -> bool:
        return self._trace_dbg.can_continue_forward

    def continue_forward(self):
        if self.can_continue_forward:
            self._sync_breakpoints()
            self._trace_dbg.continue_forward()
            self._on_state_change()

    @property
    def can_halt(self) -> bool:
        return False  # XXX: Trace playback is "instantaneous", always is halted state.

    @property
    def is_halted(self) -> bool:
        return True

    @property
    def can_stop(self) -> bool:
        return True

    def stop(self):
        pass

    @property
    def is_exited(self) -> bool:
        return False

    def replay_to_nth_event(self, n: int):
        """
        Replay to the Nth event, skipping over stop events and ending on the nearest execution event.
        """
        t = self._btrace
        assert 0 <= n < t.get_num_events()

        until = t.get_nth_event(n)
        if until is None:
            _l.error("Could not seek to event %d", n)
            return

        if self._trace_dbg.single_step_range is None:
            step_region_addr, step_region_size = None, 1
        else:
            self._trace_dbg.single_step_range: Tuple[int, int]
            step_region_addr, step_region_size = self._trace_dbg.single_step_range

        until = t.get_prev_exec_event(until, addr=step_region_addr, size=step_region_size)
        if until is None:
            _l.error("No execution event prior to event %d", n)
            return

        self._trace_dbg.state = t.replay(self._trace_dbg.state, until)
        self._on_state_change()

    def get_current_function(self):
        if self._trace_dbg.state is None:
            return None
        else:
            return self.get_function_for_event(self._trace_dbg.state.event)

    def replay_to_event(self, until):
        self._trace_dbg.state = self._btrace.replay(self._trace_dbg.state, until)
        self._on_state_change()

    #
    # Trace Analysis
    #
    # FIXME: Factor this out of debugger
    #

    def get_function_for_event(self, event: TraceEvent) -> Optional["Function"]:
        """
        Find currently execution function at `event`.
        """
        # Rewind to last block event
        if event and not isinstance(event, bintrace.FBBlockEvent):
            event = self._btrace.get_prev_bb_event(event, vcpu=self._trace_dbg.vcpu)
        if event is None:
            return None

        # Determine what function we are in currently.
        node = self.workspace.main_instance.cfg.get_node(event.Addr())
        if node is None:
            return None

        kb = self.workspace.main_instance.project.kb
        if node.function_address in kb.functions:
            return kb.functions[node.function_address], event
        else:
            _l.warning("Node %s not found in functions db", node)
            return None

    def get_called_functions(
        self, event: Optional[TraceEvent] = None, only_after_event: bool = False
    ) -> Sequence[Tuple["Function", TraceEvent]]:
        """
        Enumerate 1st order outgoing calls of function at `event`.
        """
        if event is None:
            if self._trace_dbg.state:
                event = self._trace_dbg.state.event
            else:
                return []

        # Get current function
        func = self.get_function_for_event(event)
        if func is None:
            _l.warning("Could not determine function for event %s", event)
            return []

        func, event = func
        _l.debug("Function for event %s: %s", event, func.name)

        if not only_after_event:
            # Rewind to function entry
            # FIXME: Does not properly handle nested calls to this function!
            while event.Addr() != func.addr:
                event = self._btrace.get_prev_bb_event(event, vcpu=self._trace_dbg.vcpu)
                if event is None:
                    _l.error("Did not find start of function %s (%#x) in trace", func.name, func.addr)
                    return []

        called_addrs = []
        keep_looking = True
        while keep_looking:
            # Step until next function exit
            called_func_entry_event = self._btrace.get_next_bb_event(event, vcpu=self._trace_dbg.vcpu)
            if called_func_entry_event is None:
                # End of trace
                break

            # FIXME: Possible failure in func.block_addrs_set not matching
            addr = called_func_entry_event.Addr()
            is_a_function_exit = (addr not in func.block_addrs_set) or (addr == func.addr)
            if not is_a_function_exit:
                event = called_func_entry_event
                continue

            # Check exit type
            exit_block_addr = event.Addr()
            b = self.workspace.main_instance.project.factory.block(exit_block_addr)
            if b.vex.jumpkind == "Ijk_Ret":
                _l.debug("Exit is a return to caller")
                break

            called_addrs.append((addr, called_func_entry_event))

            if b.vex.jumpkind != "Ijk_Call":
                _l.debug("Exit is a tail-call")
                break

            # FIXME: fallthru might indicate trace vs cfg inconsistency
            # Seek through events to find return site of this call
            event = called_func_entry_event
            ret_addr = b.instruction_addrs[0] + b.size
            num_nested_calls = 0
            if self.workspace.main_instance.project.arch.name == "AMD64":
                # FIXME Remove this hardcoding
                stack_reg = 7
                expected_sp = called_func_entry_event.Regs(stack_reg) + 8
            else:
                raise AssertionError("FIXME: Stack pointer check for non-x86_64")

            _l.debug("Seeking to return site for call...")
            while True:
                event = self._btrace.get_next_exec_event(event, addr=ret_addr, vcpu=self._trace_dbg.vcpu)
                if event is None:
                    _l.error(
                        "Unexpected end of trace while looking for return site in %s @ %#x "
                        "(call may have caused termination)",
                        func.name,
                        ret_addr,
                    )
                    keep_looking = False
                    break

                # Check the stack pointer at return site to ensure it matches target call
                # and does not actually belong to a nested call
                bb_event = self._btrace.get_prev_bb_event(event, vcpu=self._trace_dbg.vcpu)
                if bb_event.Regs(stack_reg) == expected_sp:
                    _l.debug("Found return block at event %s", bb_event)
                    event = bb_event
                    break
                _l.debug("Skipping over nested call (%d)", num_nested_calls)
                num_nested_calls += 1

        all_funcs = self.workspace.main_instance.project.kb.functions
        return [((all_funcs[addr] if (addr in all_funcs) else addr), e) for (addr, e) in called_addrs]

    def get_called_functions_recursive(
        self, event: Optional[TraceEvent] = None, max_depth: Optional[int] = None, depth: int = 0
    ):
        if max_depth is not None and max_depth == depth:
            return
        for func_or_addr, sub_ev in self.get_called_functions(event):
            yield func_or_addr, sub_ev, depth
            if not isinstance(func_or_addr, int):
                yield from self.get_called_functions_recursive(sub_ev, max_depth=max_depth, depth=(depth + 1))



# pylint:disable=no-self-use

# https://stackoverflow.com/questions/4795757/is-there-a-better-way-to-wordwrap-text-in-qtooltip-than-just-using-regexp
# credit: Cecil Curry

import html

from PySide6.QtCore import QEvent, QObject, Qt
from PySide6.QtWidgets import QWidget


class QAwesomeTooltipEventFilter(QObject):
    """
    Tooltip-specific event filter dramatically improving the tooltips of all
    widgets for which this filter is installed.

    Motivation
    ----------
    **Rich text tooltips** (i.e., tooltips containing one or more HTML-like
    tags) are implicitly wrapped by Qt to the width of their parent windows and
    hence typically behave as expected.

    **Plaintext tooltips** (i.e., tooltips containing no such tags), however,
    are not. For unclear reasons, plaintext tooltips are implicitly truncated to
    the width of their parent windows. The only means of circumventing this
    obscure constraint is to manually inject newlines at the appropriate
    80-character boundaries of such tooltips -- which has the distinct
    disadvantage of failing to scale to edge-case display and device
    environments (e.g., high-DPI). Such tooltips *cannot* be guaranteed to be
    legible in the general case and hence are blatantly broken under *all* Qt
    versions to date. This is a `well-known long-standing issue <issue_>`__ for
    which no official resolution exists.

    This filter globally addresses this issue by implicitly converting *all*
    intercepted plaintext tooltips into rich text tooltips in a general-purpose
    manner, thus wrapping the former exactly like the latter. To do so, this
    filter (in order):

    #. Auto-detects whether the:

       * Current event is a :class:`QEvent.ToolTipChange` event.
       * Current widget has a **non-empty plaintext tooltip**.

    #. When these conditions are satisfied:

       #. Escapes all HTML syntax in this tooltip (e.g., converting all ``&``
          characters to ``&amp;`` substrings).
       #. Embeds this tooltip in the Qt-specific ``<qt>...</qt>`` tag, thus
          implicitly converting this plaintext tooltip into a rich text tooltip.

    .. _issue:
        https://bugreports.qt.io/browse/QTBUG-41051
    """

    def eventFilter(self, widget: QObject, event: QEvent) -> bool:
        """
        Tooltip-specific event filter handling the passed Qt object and event.
        """

        # If this is a tooltip event...
        if event.type() == QEvent.ToolTipChange:
            # If the target Qt object containing this tooltip is *NOT* a widget,
            # raise a human-readable exception. While this should *NEVER* be the
            # case, edge cases are edge cases because they sometimes happen.
            if not isinstance(widget, QWidget):
                raise ValueError(f'QObject "{widget}" not a widget.')

            # Tooltip for this widget if any *OR* the empty string otherwise.
            tooltip = widget.toolTip()

            # If this tooltip is both non-empty and not already rich text...
            if tooltip and not Qt.mightBeRichText(tooltip):
                # Convert this plaintext tooltip into a rich text tooltip by:
                #
                # * Escaping all HTML syntax in this tooltip.
                # * Embedding this tooltip in the Qt-specific "<qt>...</qt>" tag.
                tooltip = f"<qt>{html.escape(tooltip)}</qt>"

                # Replace this widget's non-working plaintext tooltip with this
                # working rich text tooltip.
                widget.setToolTip(tooltip)

                # Notify the parent event handler this event has been handled.
                return True

        # Else, defer to the default superclass handling of this event.
        return super().eventFilter(widget, event)
